/*
 *  Helpers for dealing with type specifiers.
 *  Copyright © 2011, 2021 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>

#include "cdecl.h"
#include "cdecl-internal.h"

/*
 * We can represent type specifiers as a bitmap, which gives us a finite
 * list of acceptable bitmap values according to the C standard.  However,
 * the "long" specifier is allowed to occur more than once, but only at most
 * 2 times.  Treat it as a special case, assigning an unused bit to represent
 * the second long.
 */
#define CDECL_TYPE_LLONG (CDECL_SPEC_TYPE + 31)

static inline unsigned long spec_bit(unsigned type)
{
	return 1ul << (type & (CDECL_SPEC_TYPE - 1));
}

static unsigned long add_typespec(unsigned long map, struct cdecl_declspec *s)
{
	assert(s->type >= CDECL_SPEC_TYPE && s->type < CDECL_TYPE_LLONG);

	if (s->type == CDECL_TYPE_LONG) {
		if (map & spec_bit(CDECL_TYPE_LLONG)) {
			fprintf(stderr, "too many long specifiers\n");
			return -1;
		} else if (map & spec_bit(CDECL_TYPE_LONG)) {
			return map | spec_bit(CDECL_TYPE_LLONG);
		}
	}

	if (map & spec_bit(s->type)) {
		fprintf(stderr, "duplicate type specifier\n");
		return -1;
	}

	return map | spec_bit(s->type);
}

unsigned long cdecl__build_typemap(struct cdecl_declspec *s)
{
	unsigned long map = 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		if (cdecl_spec_kind(c) != CDECL_SPEC_TYPE)
			continue;

		map = add_typespec(map, c);
		if (map == -1)
			return -1;
	}

	switch (map) {
	case 0:
		fprintf(stderr, "no type specifiers\n");
		return -1;
#	include "validtypes.h"
		return map;
	default:
		fprintf(stderr, "conflicting type specifiers\n");
		return -1;
	}
}
